import logging
import os
import copy
import importlib
import openai

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.instrumentation import instrument_method, patch_method, parse_semver, compare_semver
from graphsignal import client

logger = logging.getLogger('graphsignal')

class OpenAIRecorder(BaseRecorder):
    def __init__(self):
        self._library_version = None
        self._base_url = None
        self._cached_encodings = {}
        self._extra_content_tokens = {
            'gpt-3.5-turbo': 3,
            'gpt-3.5-turbo-0301': 4,
            'gpt-4': 3,
            'gpt-4-0314': 3,
            'gpt-4-32k-0314': 3,
            'gpt-3.5-turbo-0613': 3,
            'gpt-3.5-turbo-16k-0613': 3,
            'gpt-4-0613': 3,
            'gpt-4-32k-0613': 3
        }
        self._extra_name_tokens = {
            'gpt-3.5-turbo': 1,
            'gpt-3.5-turbo-0301': -1,
            'gpt-4': 1,
            'gpt-4-0314': 1,
            'gpt-4-32k-0314': 1,
            'gpt-3.5-turbo-0613': 1,
            'gpt-3.5-turbo-16k-0613': 1,
            'gpt-4-0613': 1,
            'gpt-4-32k-0613': 1
        }
        self._extra_reply_tokens = {
            'gpt-3.5-turbo': 3,
            'gpt-3.5-turbo-0301': 3,
            'gpt-4': 3,
            'gpt-4-0314': 3,
            'gpt-4-32k-0314': 3,
            'gpt-3.5-turbo-0613': 3,
            'gpt-3.5-turbo-16k-0613': 3,
            'gpt-4-0613': 3,
            'gpt-4-32k-0613': 3
        }

    def setup(self):
        if not graphsignal._tracer.auto_instrument:
            return

        version = openai.version.VERSION
        self._library_version = version
        parsed_version = parse_semver(version)

        if compare_semver(parsed_version, (1, 0, 0)) >= 0:
            def after_init(args, kwargs, ret, exc, context):
                client = args[0]
                self._base_url = client.base_url

                instrument_method(client.chat.completions, 'create', 'openai.chat.completions.create', trace_func=self.trace_chat_completion, data_func=self.trace_chat_completion_data)
                instrument_method(client.embeddings, 'create', 'openai.embeddings.create', trace_func=self.trace_embedding)
                instrument_method(client.images, 'create', 'openai.images.create', trace_func=self.trace_image_generation)
                instrument_method(client.images, 'create_variation', 'openai.images.create_variation', trace_func=self.trace_image_variation)
                instrument_method(client.images, 'edit', 'openai.images.edit', trace_func=self.trace_image_edit)
                instrument_method(client.audio.transcriptions, 'create', 'openai.audio.transcriptions.create', trace_func=self.trace_audio_transcription)
                instrument_method(client.audio.translations, 'create', 'openai.audio.translations.create', trace_func=self.trace_audio_translation)
                instrument_method(client.moderations, 'create', 'openai.moderations.create', trace_func=self.trace_moderation)

            patch_method(openai.OpenAI, '__init__', after_func=after_init)
            patch_method(openai.AsyncOpenAI, '__init__', after_func=after_init)
            patch_method(openai.AzureOpenAI, '__init__', after_func=after_init)
            patch_method(openai.AsyncAzureOpenAI, '__init__', after_func=after_init)
        else:
            logger.debug('OpenAI tracing is only supported for >= 1.0.0.')
            return

    def shutdown(self):
        pass

    def count_tokens(self, model, text):
        if model not in self._cached_encodings:
            try:
                tiktoken = importlib.import_module('tiktoken')
                encoding = tiktoken.encoding_for_model(model)
                self._cached_encodings[model] = encoding
                if encoding:
                    logger.debug('Cached encoding for model %s', model)
                else:
                    logger.debug('No encoding returned for model %s', model)
            except ModuleNotFoundError:
                self._cached_encodings[model] = None
                logger.debug('tiktoken not installed, will not count OpenAI stream tokens.')
            except Exception:
                self._cached_encodings[model] = None
                logger.error('Error using tiktoken for model %s', model, exc_info=True)

        encoding = self._cached_encodings.get(model, None)
        if encoding:
            return len(encoding.encode(text))
        return None

    def set_common_tags(self, span, endpoint, params):
        span.set_param('openai_version', self._library_version)
        if 'openai.com' in str(self._base_url):
            span.set_tag('api_provider', 'openai')
        if 'azure.com' in str(self._base_url):
            span.set_tag('api_provider', 'azure')
        span.set_tag('endpoint', f'{self._base_url}{endpoint}')

        # header tags
        try:
            tags_header_value = None
            if 'extra_headers' in params and isinstance(params['extra_headers'], dict):
                for header_key, header_value in params['extra_headers'].items():
                    if header_key.lower() == 'graphsignal-tags':
                        tags_header_value = header_value
                        del params['extra_headers'][header_key]
                        break

            if isinstance(tags_header_value, str):
                header_tags = dict([el.strip(' ') for el in kv.split('=')] for kv in tags_header_value.split(','))
                for tag_key, tag_value in header_tags.items():
                    span.set_tag(tag_key, tag_value)
        except Exception:
            logger.error('Error parsing Graphsignal-Tags extra header', exc_info=True)

        # user tag
        if 'user' in params:
            if not graphsignal.get_tag('user_id') and not graphsignal.get_context_tag('user_id') and not span.get_tag('user_id'):
                span.set_tag('user_id', params['user'])

    def read_usage(self, span, usage):
        if not usage:
            return

        def set_usage(span, obj, key, new_key=None):
            if not isinstance(obj, dict):
                return 0

            if key in obj:
                value = obj[key]
                if isinstance(value, int):
                    if new_key:
                        span.set_counter(new_key, value)
                        span.inc_counter_metric('usage', new_key, value)
                    else:
                        span.set_counter(key, value)
                        span.inc_counter_metric('usage', key, value)
                    return value
            return None

        set_usage(span, usage, 'total_tokens')
        prompt_tokens = set_usage(span, usage, 'prompt_tokens')
        cached_prompt_tokens = None
        set_usage(span, usage, 'completion_tokens')
        if 'completion_tokens' in usage:
            span.set_counter('output_tokens', usage['completion_tokens'])
        if 'prompt_tokens_details' in usage:
            prompt_tokens_details = usage['prompt_tokens_details']
            set_usage(span, prompt_tokens_details, 'audio_tokens', 'prompt_audio_tokens')
            cached_prompt_tokens = set_usage(span, prompt_tokens_details, 'cached_tokens', 'cached_prompt_tokens')
        if 'completion_tokens_details' in usage:
            completion_tokens_details = usage['completion_tokens_details']
            set_usage(span, completion_tokens_details, 'audio_tokens', 'completion_audio_tokens')
            set_usage(span, completion_tokens_details, 'reasoning_tokens')
            set_usage(span, completion_tokens_details, 'accepted_prediction_tokens')
            set_usage(span, completion_tokens_details, 'rejected_prediction_tokens')

        if prompt_tokens:
            if cached_prompt_tokens and cached_prompt_tokens > 0:
                uncached_prompt_tokens = prompt_tokens - cached_prompt_tokens
                span.set_counter('uncached_prompt_tokens', uncached_prompt_tokens)
                span.inc_counter_metric('usage', 'uncached_prompt_tokens', uncached_prompt_tokens)
            else:
                span.set_counter('uncached_prompt_tokens', prompt_tokens)
                span.inc_counter_metric('usage', 'uncached_prompt_tokens', prompt_tokens)

    def trace_chat_completion(self, span, args, kwargs, ret, exc):
        params = kwargs

        span.set_tag('model_type', 'chat')
        self.set_common_tags(span, 'chat/completions', params)
        if 'model' in params:
            span.set_tag('model', params['model'])
            span.set_param('model', params['model'])
        if 'reasoning_effort' in params:
            span.set_param('reasoning_effort', params['reasoning_effort'])
        if 'max_completion_tokens' in params:
            span.set_param('max_completion_tokens', params['max_completion_tokens'])

        if 'stream' in params and params['stream']:
            if 'messages' in params:
                if 'model' in params and not exc:
                    # Based on token counting example
                    # https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
                    model = params['model']
                    total_prompt_tokens = 0
                    for message in params['messages']:
                        if 'content' in message:
                            prompt_tokens = self.count_tokens(model, message['content'])
                            if prompt_tokens:
                                total_prompt_tokens += prompt_tokens
                                total_prompt_tokens += self._extra_content_tokens.get(model, 0)
                        if 'role' in message:
                            prompt_tokens = self.count_tokens(model, message['role'])
                            if prompt_tokens:
                                total_prompt_tokens += prompt_tokens
                        if 'name' in message:
                            prompt_tokens = self.count_tokens(model, message['name'])
                            if prompt_tokens:
                                total_prompt_tokens += prompt_tokens
                                total_prompt_tokens += self._extra_name_tokens.get(model, 0)
                    if total_prompt_tokens > 0:
                        total_prompt_tokens += self._extra_reply_tokens.get(model, 0)
                    span.set_counter('prompt_tokens', total_prompt_tokens)
                    span.set_counter('uncached_prompt_tokens', total_prompt_tokens)
                    span.inc_counter_metric('usage', 'prompt_tokens', total_prompt_tokens)
                    span.inc_counter_metric('usage', 'uncached_prompt_tokens', total_prompt_tokens)
        else:
            ret = ret.model_dump()

            if ret and 'model' in ret:
                span.set_tag('effective_model', ret['model'])

            if ret and 'usage' in ret and not exc:
                self.read_usage(span, ret['usage'])

    def trace_chat_completion_data(self, span, item):
        item = item.model_dump()

        span.set_perf_counter('first_token_ns')
        if item and 'choices' in item:
            for _ in item['choices']:
                span.inc_counter('completion_tokens', 1)
                span.inc_counter_metric('usage', 'completion_tokens', 1)
                span.inc_counter('output_tokens', 1)

        if item and 'usage' in item:
            self.read_usage(span, item['usage'])

    def trace_embedding(self, span, args, kwargs, ret, exc):
        params = kwargs

        self.set_common_tags(span, 'embeddings', params)
        if 'model' in params:
            span.set_tag('model', params['model'])
            span.set_param('model', params['model'])

        ret = ret.model_dump()

        if ret and 'model' in ret:
            span.set_tag('effective_model', ret['model'])

        if ret and 'usage' in ret:
            usage = ret['usage']
            if 'total_tokens' in usage:
                span.set_counter('total_tokens', usage['total_tokens'])
                span.inc_counter_metric('usage', 'total_tokens', usage['total_tokens'])
            if 'prompt_tokens' in usage:
                span.set_counter('prompt_tokens', usage['prompt_tokens'])
                span.inc_counter_metric('usage', 'prompt_tokens', usage['prompt_tokens'])

    def trace_image_generation(self, span, args, kwargs, ret, exc):
        params = kwargs

        self.set_common_tags(span, 'images/generations', params)

        self.trace_image_endpoint(span, args, kwargs, ret, exc)

    def trace_image_variation(self, span, args, kwargs, ret, exc):
        params = kwargs

        self.set_common_tags(span, 'images/variations', params)

        self.trace_image_endpoint(span, args, kwargs, ret, exc)

    def trace_image_edit(self, span, args, kwargs, ret, exc):
        params = kwargs

        self.set_common_tags(span, 'images/edits', params)

        self.trace_image_endpoint(span, args, kwargs, ret, exc)

    def trace_image_endpoint(self, span, args, kwargs, ret, exc):
        params = kwargs

        if ret and 'data' in ret:
            image_data = []
            for image in ret['data']:
                if 'url' in image:
                    image_data.append(image['url'])
                elif 'b64_json' in image:
                    image_data.append(image['b64_json'])

    def trace_audio_transcription(self, span, args, kwargs, ret, exc):
        params = kwargs

        self.set_common_tags(span, 'audio/transcriptions', params)
        if 'model' in params:
            span.set_tag('model', params['model'])
            span.set_param('model', params['model'])

    def trace_audio_translation(self, span, args, kwargs, ret, exc):
        params = kwargs

        self.set_common_tags(span, 'audio/translations', params)
        if 'model' in params:
            span.set_tag('model', params['model'])
            span.set_param('model', params['model'])

    def trace_moderation(self, span, args, kwargs, ret, exc):
        params = kwargs

        self.set_common_tags(span, 'moderations', params)
        if 'model' in params:
            span.set_tag('model', params['model'])
            span.set_param('model', params['model'])



import logging
import os
import json
import time
import torch

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.profiler_utils import create_log_dir, remove_log_dir

logger = logging.getLogger('graphsignal')

class PyTorchRecorder(BaseRecorder):
    def __init__(self):
        self._torch_prof = None
        self._log_dir = None

    def setup(self):
        pass

    def on_span_start(self, span, context):
        if not span._with_profile:
            return
        if self._torch_prof:
            # profiler active, skip
            # only one profiler is allowed per process, so we don't need global locks
            return
        if not span.profiled():
            return

        if not self._torch_prof:
            def _schedule_func(step):
                return torch.profiler.ProfilerAction.RECORD
            self._torch_prof = torch.profiler.profile(
                schedule=_schedule_func,
                record_shapes=False,
                profile_memory=True,
                with_stack=False,
                with_flops=True)

        self._torch_prof.start()

    def on_span_stop(self, span, context):
        if not self._torch_prof:
            return

        self._torch_prof.stop()

    def on_span_read(self, span, context):
        if not self._torch_prof:
            return
        
        span.set_param('pytorch_version', torch.__version__)

        try:
            cpu_profile = []
            for event_avg in self._torch_prof.key_averages():
                if event_avg.key and event_avg.key.startswith('ProfilerStep'):
                    continue
                cpu_profile.append(dict(
                    op_name = event_avg.key,
                    device_type = event_avg.device_type.name if event_avg.device_type else None,
                    count = _uint(event_avg.count),
                    cpu_time_ns = _ns(event_avg.cpu_time_total),
                    self_cpu_time_ns = _ns(event_avg.self_cpu_time_total),
                    device_time_ns = _ns(event_avg.device_time_total),
                    self_device_time_ns = _ns(event_avg.self_device_time_total),
                    cpu_memory = _uint(event_avg.cpu_memory_usage),
                    self_cpu_memory = _uint(event_avg.self_cpu_memory_usage),
                    device_memory = _uint(event_avg.device_memory_usage),
                    self_device_memory = _uint(event_avg.self_device_memory_usage),
                    flops = _uint(event_avg.flops)
                ))

            if len(cpu_profile) > 0:
                span.set_profile('cpu-profile', 'event-averages', json.dumps(cpu_profile))
                span.set_tag('profile_type', 'cpu')

            kernel_index = {}
            for event in self._torch_prof.events():
                for kernel in event.kernels:
                    key = (event.key, kernel.name, kernel.device)
                    if key in kernel_index:
                        kernel_avg = kernel_index[key]
                        kernel_avg['count'] += 1
                        kernel_avg['duration_ns'] += _ns(kernel.duration)
                    else:
                        kernel_avg = kernel_index[key] = dict(
                            device_idx = kernel.device,
                            op_name = event.name,
                            kernel_name = kernel.name,
                            count = 1,
                            duration_ns = _ns(kernel.duration)
                        )
            
            device_profile = kernel_index.values()
            if len(device_profile) > 0:
                span.set_profile('device-profile', 'event-averages', json.dumps(device_profile))
                span.set_tag('profile_type', 'device') # override cpu value

            chrome_trace = self._export_chrome_trace()
            if chrome_trace:
                span.set_profile('event-timeline', 'chrome-trace', chrome_trace)

            if len(cpu_profile) > 0 or len(device_profile) > 0:
                span.set_param('profiler', f'pytorch-{torch.__version__}')
        finally:
            self._torch_prof = None

    def _export_chrome_trace(self):
        try:
            read_start_time = time.time()

            self._log_dir = create_log_dir()

            trace_path = os.path.join(self._log_dir, 'trace.json')
            self._torch_prof.export_chrome_trace(trace_path)

            trace_file_size = os.path.getsize(trace_path)
            logger.debug('Chrome trace size: %s', trace_file_size)
            if trace_file_size > 50 * 1e6:
                raise Exception('Trace file too big: {0}'.format(trace_file_size))

            with open(trace_path, "r") as f:
                return str(f.read())
        finally:
            remove_log_dir(self._log_dir)
            logger.debug('Chrome trace export time: %s', time.time() - read_start_time)

        return None

def _ns(val):
    return int(max(val, 0) * 1e3)


def _uint(val):
    return max(int(val), 0)



# coding: utf-8

"""
    Graphsignal API

    API for uploading and querying spans, scores, metrics, and logs.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from graphsignal.client.models.tag import Tag
from typing import Optional, Set
from typing_extensions import Self

class Score(BaseModel):
    """
    Score
    """ # noqa: E501
    score_id: Optional[StrictStr] = Field(default=None, description="Unique identifier for the score.")
    span_id: Optional[StrictStr] = Field(default=None, description="The associated span identifier, if the score is being associated with a span.")
    tags: Optional[List[Tag]] = Field(default=None, description="Tags associated with the score.")
    name: StrictStr = Field(description="The name of the score.")
    comment: Optional[StrictStr] = Field(default=None, description="Optional comment or description for the score.")
    score: Union[StrictFloat, StrictInt] = Field(description="The numerical value of the score.")
    unit: Optional[StrictStr] = Field(default=None, description="The units of score value.")
    severity: Optional[StrictInt] = Field(default=None, description="Severity of the score, 1-5 (info, low, medium, high, critical).")
    create_ts: StrictInt = Field(description="Unix timestamp (seconds) when the score was created.")
    __properties: ClassVar[List[str]] = ["score_id", "span_id", "tags", "name", "comment", "score", "unit", "severity", "create_ts"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of Score from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in tags (list)
        _items = []
        if self.tags:
            for _item_tags in self.tags:
                if _item_tags:
                    _items.append(_item_tags.to_dict())
            _dict['tags'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of Score from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "score_id": obj.get("score_id"),
            "span_id": obj.get("span_id"),
            "tags": [Tag.from_dict(_item) for _item in obj["tags"]] if obj.get("tags") is not None else None,
            "name": obj.get("name"),
            "comment": obj.get("comment"),
            "score": obj.get("score"),
            "unit": obj.get("unit"),
            "severity": obj.get("severity"),
            "create_ts": obj.get("create_ts")
        })
        return _obj





from typing import Dict, Any, Union, Optional, Type
import os
import logging
import atexit

from graphsignal.version import __version__
from graphsignal.tracer import Tracer
from graphsignal.spans import Span, SpanContext

logger = logging.getLogger('graphsignal')

_tracer = None


def _check_configured():
    global _tracer
    if not _tracer:
        raise ValueError(
            'Tracer not configured, call graphsignal.configure() first')


def _parse_env_param(name: str, value: Any, expected_type: Type) -> Any:
    if value is None:
        return None

    try:
        if expected_type == bool:
            return value if isinstance(value, bool) else str(value).lower() in ("true", "1", "yes")
        elif expected_type == int:
            return int(value)
        elif expected_type == float:
            return float(value)
        elif expected_type == str:
            return str(value)
    except (ValueError, TypeError):
        pass

    raise ValueError(f"Invalid type for {name}: expected {expected_type.__name__}, got {type(value).__name__}")


def _read_config_param(name: str, expected_type: Type, provided_value: Optional[Any] = None, default_value: Optional[Any] = None, required: bool = False) -> Any:
    # Check if the value was provided as an argument
    if provided_value is not None:
        return provided_value

    # Check if the value was provided as an environment variable
    env_value = os.getenv(f'GRAPHSIGNAL_{name.upper()}')
    if env_value is not None:
        parsed_env_value = _parse_env_param(name, env_value, expected_type)
        if parsed_env_value is not None:
            return parsed_env_value

    if required:
        raise ValueError(f"Missing required argument: {name}")

    return default_value


def _read_config_tags(provided_value: Optional[dict] = None, prefix: str = "GRAPHSIGNAL_TAG_") -> Dict[str, str]:
    # Check if the value was provided as an argument
    if provided_value is not None:
        return provided_value

    # Check if the value was provided as an environment variable
    return {key[len(prefix):].lower(): value for key, value in os.environ.items() if key.startswith(prefix)}


def configure(
    api_key: Optional[str] = None,
    api_url: Optional[str] = None,
    deployment: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    auto_instrument: Optional[bool] = None,
    profiling_rate: Optional[float] = None,
    debug_mode: Optional[bool] = None
) -> None:
    global _tracer

    if _tracer:
        logger.warning("Tracer already configured")
        return

    api_key = _read_config_param("api_key", str, api_key, required=True)
    api_url = _read_config_param("api_url", str, api_url)
    tags = _read_config_tags(tags)
    auto_instrument = _read_config_param("auto_instrument", bool, auto_instrument, default_value=True)
    profiling_rate = _read_config_param("profiling_rate", float, profiling_rate, default_value=0.1)
    debug_mode = _read_config_param("debug_mode", bool, debug_mode, default_value=False)

    # left for compatibility
    if deployment and isinstance(deployment, str):
        tags['deployment'] = deployment

    _tracer = Tracer(
        api_key=api_key,
        api_url=api_url,
        tags=tags,
        auto_instrument=auto_instrument,
        profiling_rate=profiling_rate,
        debug_mode=debug_mode)
    _tracer.setup()

    atexit.register(shutdown)

    logger.debug('Tracer configured')


def set_tag(key: str, value: str) -> None:
    _check_configured()
    _tracer.set_tag(key, value)


def get_tag(key: str) -> Optional[str]:
    _check_configured()
    return _tracer.get_tag(key)


def remove_tag(key: str):
    _check_configured()
    return _tracer.remove_tag(key)


def set_context_tag(key: str, value: str, append_uuid=None) -> None:
    _check_configured()
    _tracer.set_context_tag(key, value, append_uuid=append_uuid)


def get_context_tag(key: str) -> Optional[str]:
    _check_configured()
    return _tracer.get_context_tag(key)


def remove_context_tag(key: str):
    _check_configured()
    return _tracer.remove_context_tag(key)


def set_param(name: str, value: Any) -> None:
    _check_configured()
    _tracer.set_param(name, value)


def get_param(name: str) -> Optional[Any]:
    _check_configured()
    return _tracer.get_param(name)


def remove_param(name: str):
    _check_configured()
    return _tracer.remove_param(name)


def trace(
        operation: str,
        tags: Optional[Dict[str, str]] = None,
        with_profile: Optional[bool] = False) -> 'Span':
    _check_configured()

    return _tracer.trace(operation=operation, tags=tags, with_profile=with_profile)


def trace_function(
        func=None, 
        *,
        operation: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        with_profile: Optional[bool] = False) -> Any:
    return _tracer.trace_function(func, operation=operation, tags=tags, with_profile=with_profile)


def score(
        name: str, 
        tags: Optional[Dict[str, str]] = None,
        score: Optional[Union[int, float]] = None, 
        unit: Optional[str] = None,
        severity: Optional[int] = None,
        comment: Optional[str] = None) -> None:
    _check_configured()
    return _tracer.score(
        name=name, 
        tags=tags, 
        score=score, 
        unit=unit, 
        severity=severity, 
        comment=comment)


def upload(block=False) -> None:
    _check_configured()

    _tracer.upload(block=block)


def shutdown() -> None:
    global _tracer
    if not _tracer:
        return

    atexit.unregister(shutdown)
    _tracer.shutdown()
    _tracer = None

    logger.debug('Tracer shutdown')


__all__ = [
    '__version__',
    'configure',
    'upload',
    'shutdown',
    'trace',
    'function_trace',
    'SpanContext',
    'Span',
    'score',
    'set_tag',
    'get_tag',
    'set_context_tag',
    'get_context_tag',
    'remove_context_tag'
]
